"""Main."""

import asyncio
import logging

from wipac_dev_tools import logging_tools

from .config import ENV
from .pilot import consume_and_reply

LOGGER = logging.getLogger(__package__)


def main() -> None:
    """Start up EWMS Pilot to do tasks, communicate via message passing."""

    logging_tools.set_level(
        ENV.EWMS_PILOT_CL_LOG,  # type: ignore[arg-type]
        first_party_loggers=[LOGGER],
        third_party_level=ENV.EWMS_PILOT_CL_LOG_THIRD_PARTY,  # type: ignore[arg-type]
        use_coloredlogs=True,
    )

    # GO!
    LOGGER.info("Running from CL: starting up an EWMS Pilot...")
    asyncio.run(consume_and_reply())
    LOGGER.info("Done.")


if __name__ == "__main__":
    main()



"""Logic for waiting on task set."""

import asyncio
import logging
from typing import Dict, List, Set, Tuple

import mqclient as mq
from mqclient.broker_client_interface import Message

from ..utils.utils import all_task_errors_string

LOGGER = logging.getLogger(__name__)

AsyncioTaskMessages = Dict[asyncio.Task, Message]  # type: ignore[type-arg]


async def wait_on_tasks_with_ack(
    sub: mq.queue.ManualQueueSubResource,
    pub: mq.queue.QueuePubResource,
    tasks_msgs: AsyncioTaskMessages,
    previous_task_errors: List[BaseException],
    timeout: int,
) -> Tuple[AsyncioTaskMessages, List[BaseException]]:
    """Get finished tasks and ack/nack their messages.

    Returns:
        Tuple:
            AsyncioTaskMessages: pending tasks and
            List[BaseException]: failed tasks' exceptions (plus those in `previous_task_errors`)
    """
    pending: Set[asyncio.Task] = set(tasks_msgs.keys())  # type: ignore[type-arg]
    if not pending:
        return {}, previous_task_errors

    async def handle_failed_task(task: asyncio.Task, exception: BaseException) -> None:  # type: ignore[type-arg]
        previous_task_errors.append(exception)
        LOGGER.error(
            f"TASK FAILED ({repr(exception)}) -- attempting to nack original message..."
        )
        try:
            await sub.nack(tasks_msgs[task])
        except Exception as e:
            # LOGGER.exception(e)
            LOGGER.error(f"Could not nack: {repr(e)}")
        LOGGER.error(all_task_errors_string(previous_task_errors))

    # wait for next task
    LOGGER.debug("Waiting on tasks...")
    done, pending = await asyncio.wait(
        pending,
        return_when=asyncio.FIRST_COMPLETED,
        timeout=timeout,
    )

    # HANDLE FINISHED TASK(S)
    # fyi, most likely one task in here, but 2+ could finish at same time
    for task in done:
        try:
            result = await task
        except Exception as e:
            LOGGER.exception(e)
            # FAILED TASK!
            await handle_failed_task(task, e)
            continue

        # SUCCESSFUL TASK -> send result
        try:
            LOGGER.info("TASK FINISHED -- attempting to send result message...")
            await pub.send(result)
        except Exception as e:
            # -> failed to send = FAILED TASK!
            LOGGER.error(
                f"Failed to send finished task's result: {repr(e)}"
                f" -- task now considered as failed"
            )
            await handle_failed_task(task, e)
            continue

        # SUCCESSFUL TASK -> result sent -> ack original message
        try:
            LOGGER.info("Now, attempting to ack original message...")
            await sub.ack(tasks_msgs[task])
        except mq.broker_client_interface.AckException as e:
            # -> result sent -> ack failed = that's okay!
            LOGGER.error(
                f"Could not ack ({repr(e)}) -- not counting as a failed task"
                " since task's result was sent successfully -- "
                "NOTE: outgoing queue may eventually get"
                " duplicate result when original message is"
                " re-delivered by broker to another pilot"
                " & the new result is sent"
            )

    if done:
        LOGGER.info(f"{len(tasks_msgs)-len(pending)} Tasks Finished")

    return (
        {t: msg for t, msg in tasks_msgs.items() if t in pending},
        # this now also includes tasks that finished this round
        previous_task_errors,
    )



"""Logic for running a subprocess."""

import asyncio
import logging
import sys
from pathlib import Path
from typing import Optional, TextIO

from ewms_pilot.config import ENV

LOGGER = logging.getLogger(__name__)


def get_last_line(fpath: Path) -> str:
    """Get the last line of the file."""
    with fpath.open() as f:
        line = ""
        for line in f:
            pass
        return line.rstrip()  # remove trailing '\n'


class PilotSubprocessError(Exception):
    """Raised when the subprocess terminates in an error."""

    def __init__(self, return_code: int, stderrfile: Path):
        super().__init__(
            f"Subprocess completed with exit code {return_code}: "
            f"{get_last_line(stderrfile)}"
        )


def _dump_binary_file(fpath: Path, stream: TextIO) -> None:
    try:
        with open(fpath, "rb") as file:
            while True:
                chunk = file.read(4096)
                if not chunk:
                    break
                stream.buffer.write(chunk)
    except Exception as e:
        LOGGER.error(f"Error dumping subprocess output ({stream.name}): {e}")


async def run_container(
    image: str,
    args: str,
    timeout: Optional[int],
    stdoutfile: Path,
    stderrfile: Path,
    mount_bindings: str = "",
    env_options: str = "",
) -> None:
    """Run the container and dump outputs."""
    dump_output = ENV.EWMS_PILOT_DUMP_TASK_OUTPUT

    # NOTE: don't add to mount_bindings (WYSIWYG); also avoid intermediate structures
    cmd = f"docker run --rm -i {mount_bindings} {env_options} {image} {args}"
    LOGGER.info(cmd)

    # call & check outputs
    try:
        with open(stdoutfile, "wb") as stdoutf, open(stderrfile, "wb") as stderrf:
            # await to start & prep coroutines
            proc = await asyncio.create_subprocess_shell(
                cmd,
                stdout=stdoutf,
                stderr=stderrf,
            )
            # await to finish
            try:
                await asyncio.wait_for(  # raises TimeoutError
                    proc.wait(),
                    timeout=timeout,
                )
            except (TimeoutError, asyncio.exceptions.TimeoutError) as e:
                # < 3.11 -> asyncio.exceptions.TimeoutError
                raise TimeoutError(f"subprocess timed out after {timeout}s") from e

        LOGGER.info(f"Subprocess return code: {proc.returncode}")

        # exception handling (immediately re-handled by 'except' below)
        if proc.returncode:
            raise PilotSubprocessError(proc.returncode, stderrfile)

    except Exception as e:
        LOGGER.error(f"Subprocess failed: {e}")  # log the time
        dump_output = True
        raise
    finally:
        if dump_output:
            _dump_binary_file(stdoutfile, sys.stdout)
            _dump_binary_file(stderrfile, sys.stderr)



"""Configuration constants."""

import dataclasses as dc
import logging
import os
import shutil
from pathlib import Path
from typing import Optional

from wipac_dev_tools import from_environment_as_dataclass

LOGGER = logging.getLogger(__name__)


REFRESH_INTERVAL = 1  # sec -- the time between transitioning phases of the main loop


#
# Env var constants: set as constants & typecast
#


@dc.dataclass(frozen=True)
class EnvConfig:
    """For storing environment variables, typed."""

    #
    # REQUIRED
    #

    # task
    EWMS_PILOT_TASK_IMAGE: str = ""  # the image to run for each task
    EWMS_PILOT_TASK_ARGS: str = ""  # the args to run with the task container

    # init
    EWMS_PILOT_INIT_IMAGE: str = ""  # the image to run once before processing any tasks
    EWMS_PILOT_INIT_ARGS: str = ""  # the args to run with the init container

    # incoming queue
    EWMS_PILOT_QUEUE_INCOMING: str = ""  # name of the incoming queue
    EWMS_PILOT_QUEUE_INCOMING_AUTH_TOKEN: str = ""  # auth token for queue
    EWMS_PILOT_QUEUE_INCOMING_BROKER_TYPE: str = ""  # broker type: pulsar, rabbitmq...
    EWMS_PILOT_QUEUE_INCOMING_BROKER_ADDRESS: str = ""  # MQ broker URL to connect to

    # outgoing queue
    EWMS_PILOT_QUEUE_OUTGOING: str = ""  # name of the outgoing queue
    EWMS_PILOT_QUEUE_OUTGOING_AUTH_TOKEN: str = ""  # auth token for queue
    EWMS_PILOT_QUEUE_OUTGOING_BROKER_TYPE: str = ""  # broker type: pulsar, rabbitmq...
    EWMS_PILOT_QUEUE_OUTGOING_BROKER_ADDRESS: str = ""  # MQ broker URL to connect to

    #
    # OPTIONAL
    #

    EWMS_PILOT_DATA_DIR_PARENT_PATH_ON_HOST: str = ""

    # I/O to subprocess -- the file type (extension) of the input/output file from the pilot's task
    EWMS_PILOT_INFILE_TYPE: str = ".in"  # ''
    EWMS_PILOT_OUTFILE_TYPE: str = ".out"  # ''

    # incoming queue - settings
    EWMS_PILOT_PREFETCH: int = (
        1  # prefetch amount for incoming messages (off by default -- prefetch is an optimization)
    )
    EWMS_PILOT_TIMEOUT_QUEUE_WAIT_FOR_FIRST_MESSAGE: Optional[int] = (
        None  # timeout (sec) for the first message to arrive at the pilot (defaults to incoming timeout value)
    )
    EWMS_PILOT_TIMEOUT_QUEUE_INCOMING: int = 1  # timeout (sec) for messages TO pilot

    # files
    EWMS_PILOT_EXTERNAL_DIRECTORIES: str = ""  # comma-delimited

    # logging -- only used when running via command line
    EWMS_PILOT_CL_LOG: str = "INFO"  # level for 1st-party loggers
    EWMS_PILOT_CL_LOG_THIRD_PARTY: str = "WARNING"  # level for 3rd-party loggers

    # chirp
    EWMS_PILOT_HTCHIRP: bool = False
    EWMS_PILOT_HTCHIRP_DEST: str = "JOB_ATTR"  # ["JOB_EVENT_LOG", "JOB_ATTR"]
    EWMS_PILOT_HTCHIRP_RATELIMIT_INTERVAL: float = 60.0

    # timing config -- tasks
    EWMS_PILOT_INIT_TIMEOUT: Optional[int] = None  # timeout (sec) for the init command
    EWMS_PILOT_TASK_TIMEOUT: Optional[int] = None  # timeout (sec) for each task

    # task handling logic
    EWMS_PILOT_STOP_LISTENING_ON_TASK_ERROR: bool = (
        True
        # whether to stop taking future tasks after a task fails;
        # ex: set to False if on known good compute node (testing cluster),
        #     set to True  if on unknown node (large homogeneous cluster)
    )
    EWMS_PILOT_MAX_CONCURRENT_TASKS: int = 1  # max no. of tasks to process in parallel

    # misc settings
    EWMS_PILOT_KEEP_ALL_TASK_FILES: bool = False
    EWMS_PILOT_DUMP_TASK_OUTPUT: bool = (
        False  # dump each task's stderr to stderr and stdout to stdout
    )
    EWMS_PILOT_QUARANTINE_TIME: int = (
        0  # how long to sleep after error (useful for preventing blackhole scenarios on condor)
    )

    def __post_init__(self) -> None:
        if timeout := os.getenv("EWMS_PILOT_SUBPROC_TIMEOUT"):
            LOGGER.warning(
                "Using 'EWMS_PILOT_SUBPROC_TIMEOUT'; 'EWMS_PILOT_TASK_TIMEOUT' is preferred."
            )
            if self.EWMS_PILOT_TASK_TIMEOUT is not None:
                LOGGER.warning(
                    "Ignoring 'EWMS_PILOT_SUBPROC_TIMEOUT' since 'EWMS_PILOT_TASK_TIMEOUT' was provided."
                )
            else:
                # b/c frozen
                object.__setattr__(self, "EWMS_PILOT_TASK_TIMEOUT", int(timeout))

        if self.EWMS_PILOT_MAX_CONCURRENT_TASKS < 1:
            LOGGER.warning(
                f"Invalid value for 'EWMS_PILOT_MAX_CONCURRENT_TASKS' ({self.EWMS_PILOT_MAX_CONCURRENT_TASKS}),"
                " defaulting to '1'."
            )
            object.__setattr__(self, "EWMS_PILOT_CONCURRENT_TASKS", 1)  # b/c frozen

        if (
            self.EWMS_PILOT_QUARANTINE_TIME
            and not self.EWMS_PILOT_STOP_LISTENING_ON_TASK_ERROR
        ):
            raise RuntimeError(
                f"Cannot define 'EWMS_PILOT_QUARANTINE_TIME' while "
                f"'EWMS_PILOT_STOP_LISTENING_ON_TASK_ERROR' is "
                f"'{self.EWMS_PILOT_STOP_LISTENING_ON_TASK_ERROR}'"
            )

        if self.EWMS_PILOT_HTCHIRP_DEST not in ["JOB_EVENT_LOG", "JOB_ATTR"]:
            raise RuntimeError(
                f"Invalid EWMS_PILOT_HTCHIRP_DEST: {self.EWMS_PILOT_HTCHIRP_DEST}"
            )


ENV = from_environment_as_dataclass(EnvConfig)


#
# --------------------------------------------------------------------------------------
#


PILOT_DATA_DIR = Path(
    f"{ENV.EWMS_PILOT_DATA_DIR_PARENT_PATH_ON_HOST.rstrip('/')}/ewms-pilot-data"
)
PILOT_DATA_HUB_DIR = PILOT_DATA_DIR / "data-hub"

INCONTAINER_ENVNAME_TASK_DATA_HUB_DIR = "EWMS_TASK_DATA_HUB_DIR"


class DirectoryCatalog:
    """Handles the naming and mapping logic for a task's directories."""

    @dc.dataclass
    class _ContainerBindMountDirPair:
        on_host: Path
        in_container: Path

    def __init__(self, name: str):
        """Directories are not pre-created; you must `mkdir -p` to use."""
        self._namebased_dir = PILOT_DATA_DIR / name

        # for inter-task/init storage: startup data, init container's output, etc.
        self.pilot_data_hub = self._ContainerBindMountDirPair(
            PILOT_DATA_HUB_DIR,
            PILOT_DATA_HUB_DIR,
        )

        # for persisting stderr and stdout
        self.outputs_on_host = self._namebased_dir / "outputs"

        # for message-based task i/o
        self.task_io = self._ContainerBindMountDirPair(
            self._namebased_dir / "task-io",
            PILOT_DATA_DIR / "task-io",
        )

    def assemble_bind_mounts(
        self,
        external_directories: bool = False,
        task_io: bool = False,
    ) -> str:
        """Get the docker bind mount string containing the wanted directories."""
        string = f"--mount type=bind,source={self.pilot_data_hub.on_host},target={self.pilot_data_hub.in_container} "

        if external_directories:
            string += "".join(
                f"--mount type=bind,source={dpath},target={dpath},readonly "
                for dpath in ENV.EWMS_PILOT_EXTERNAL_DIRECTORIES.split(",")
            )

        if task_io:
            string += f"--mount type=bind,source={self.task_io.on_host},target={self.task_io.in_container} "

        return string

    def rm_unique_dirs(self) -> None:
        """Remove all directories (on host) created for use only by this container."""
        shutil.rmtree(self._namebased_dir)  # rm -r



"""Single task logic."""

import logging
from typing import Any, Optional

from mqclient.broker_client_interface import Message

from .io import FileExtension, InFileInterface, OutFileInterface
from ..config import (
    DirectoryCatalog,
    ENV,
    INCONTAINER_ENVNAME_TASK_DATA_HUB_DIR,
)
from ..utils.runner import run_container

LOGGER = logging.getLogger(__name__)


async def process_msg_task(
    in_msg: Message,
    #
    task_image: str,
    task_args: str,
    task_timeout: Optional[int],
    #
    infile_ext: FileExtension,
    outfile_ext: FileExtension,
) -> Any:
    """Process the message's task in a subprocess using `cmd` & respond."""

    # staging-dir logic -- includes stderr/stdout files (see below)
    dirs = DirectoryCatalog(str(in_msg.uuid))
    dirs.outputs_on_host.mkdir(parents=True, exist_ok=False)
    dirs.task_io.on_host.mkdir(parents=True, exist_ok=False)

    # create in/out file *names* -- piggy-back the uuid since it's unique and trackable
    infile_name = f"infile-{in_msg.uuid}.{infile_ext}"
    outfile_name = f"outfile-{in_msg.uuid}.{outfile_ext}"

    # insert in/out files *paths* into task_args
    task_args = task_args.replace(
        "{{INFILE}}",
        str(dirs.task_io.in_container / infile_name),
    )
    task_args = task_args.replace(
        "{{OUTFILE}}",
        str(dirs.task_io.in_container / outfile_name),
    )

    # do task
    InFileInterface.write(in_msg, dirs.task_io.on_host / infile_name)
    await run_container(
        task_image,
        task_args,
        task_timeout,
        dirs.outputs_on_host / "stderrfile",
        dirs.outputs_on_host / "stdoutfile",
        dirs.assemble_bind_mounts(external_directories=True, task_io=True),
        f"--env {INCONTAINER_ENVNAME_TASK_DATA_HUB_DIR}={dirs.pilot_data_hub.in_container}",
    )
    out_data = OutFileInterface.read(dirs.task_io.on_host / outfile_name)

    # send
    LOGGER.info("Sending response message...")

    # cleanup -- on success only
    if not ENV.EWMS_PILOT_KEEP_ALL_TASK_FILES:
        dirs.rm_unique_dirs()

    return out_data
